/* ---------------------------------------------------------------------
 * 
 * Giada - Your Hardcore Loopmachine 
 * 
 * ---------------------------------------------------------------------
 * 
 * Copyright (C) 2010-2012 Giovanni A. Zuliani | Monocasual
 * 
 * This file is part of Giada - Your Hardcore Loopmachine.
 * 
 * Giada - Your Hardcore Loopmachine is free software: you can 
 * redistribute it and/or modify it under the terms of the GNU General 
 * Public License as published by the Free Software Foundation, either 
 * version 3 of the License, or (at your option) any later version.
 * 
 * Giada - Your Hardcore Loopmachine is distributed in the hope that it 
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Giada - Your Hardcore Loopmachine. If not, see 
 * <http://www.gnu.org/licenses/>.
 * 
 * ------------------------------------------------------------------ */



#ifndef RECORDER_H
#define RECORDER_H

#include <stdio.h>
#include <stdlib.h>
#include "const.h"
#include "utils.h"
#include "mixer.h"
#include "mixerHandler.h"
#include "kernelAudio.h"



namespace recorder {

struct action {
	int  chan;
	char type;
	int  frame; 
};



struct composite {
	action a1;
	action a2;
};

extern gVector<int>  frames;					      
extern gVector< gVector<action*> > global;	
extern gVector<action*>  actions;				    

extern bool active;
extern bool chanActive[MAX_NUM_CHAN];				
extern bool chanEvents[MAX_NUM_CHAN];				
extern bool sortedActions;                  



void init();



void chanHasEvents(int chan);



bool canRec(int chan);



void rec(int chan, char action, int frame);



void clearChan(int chan);



void clearAction(int chan, char action);



void deleteAction(int chan, int frame, char type);



void deleteActions(int chan, int frame_a, int frame_b, char type);



void clearAll();



void optimize();



void sortActions();



void updateBpm(float oldval, float newval, int oldquanto);



void updateSamplerate(int systemRate, int patchRate);

void expand(int old_fpb, int new_fpb);
void shrink(int new_fpb);



void enableRead(int chan);
void disableRead(int chan);



int getStartActionFrame(int chan, char action, int frame);



int getEndActionFrame(int chan, char action, int frame);



int getNextAction(int chan, char action, int frame, struct action **out);



void startOverdub(int chan, char action, int frame);
void stopOverdub(int frame);



void print();

}  

#endif
